__version__ = '1.0'

import os
import yaml
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


class Config:
    def __init__(self, file_obj=None):
        self._cached_config = {}
        self._cast_func_mapping = {}
        if file_obj:
            self.load(file_obj)

    def load(self, file_obj):
        # file_obj may be filename
        if isinstance(file_obj, str):
            filename = file_obj
            with open(filename) as file_obj:
                config = yaml.load(file_obj, Loader=Loader)
        else:
            config = yaml.load(file_obj, Loader=Loader)
        self.merge_config(config)

    def load_env(self):
        self.merge_config(os.environ)

    def set_cast_func(self, key_str, cast_func):
        self._cast_func_mapping[key_str.upper()] = cast_func

    def _set_config(self, key_str, value):
        nested_key_lst = key_str.upper().lstrip('.').split('.')
        config_dict = self._cached_config
        for index, key in enumerate(nested_key_lst):
            if index < len(nested_key_lst) - 1:
                if not isinstance(config_dict.get(key), dict):
                    config_dict[key] = {}
                config_dict = config_dict[key]
            else:
                config_dict[key] = value

    def merge_config(self, new_config, prefix_str=''):
        for key, value in new_config.items():
            key_str = '{prefix_str}.{key}'.format(prefix_str=prefix_str,
                                                  key=key)
            if isinstance(value, dict):
                self.merge_config(value, prefix_str=key_str)
            else:
                self._set_config(key_str, value)

    def __getitem__(self, key_str):
        config = self._cached_config
        for key in key_str.upper().split('.'):
            config = config[key]
        cast_func = self._cast_func_mapping.get(key_str.upper())
        if cast_func:
            config = cast_func(config)
        return config



import os

from p_config import Config

base_dir = os.path.dirname(os.path.abspath(__file__))


def test_default():
    config = Config(os.path.join(base_dir, 'default.yml'))
    assert config['server.port'] == 80
    assert config['server.hostname'] == 'localhost'
    assert config['dbs.mysql.hostname'] == 'localhost'
    assert config['dbs.mysql.port'] == 3306
    assert config['dbs.mysql.options.charset'] == 'utf8mb4'
    assert config['hostnames'] == ['demo.com']
    assert config['nodes'] == ['127.0.0.2', '127.0.0.3']
    # test upper case
    assert config['DBS.MYSQL.OPTIONS.CHARSET'] == 'utf8mb4'
    # test title case
    assert config['Dbs.Mysql.Options.Charset'] == 'utf8mb4'

    try:
        config['server.unknow']
        raise Exception('expect KeyError')
    except KeyError:
        pass


def test_override_yaml():
    config = Config(os.path.join(base_dir, 'default.yml'))
    config.load(os.path.join(base_dir, 'override.yml'))

    assert config['server.port'] == 8080
    assert config['server.hostname'] == 'localhost'
    assert config['dbs.mysql.hostname'] == 'online.com'
    assert config['dbs.mysql.port'] == 3308
    assert config['dbs.mysql.options.charset'] == 'utf8'
    assert config['dbs.mysql.options.timeout'] == 1000
    assert config['dbs.postgresql.hostname'] == 'online.com'
    assert config['dbs.postgresql.username'] == 'root'
    assert config['nodes.node1'] == '127.0.0.1'
    assert config['nodes'] == {'NODE1': '127.0.0.1'}


def test_environ():
    config = Config()
    config.set_cast_func('server.port', int)

    os.environ['server.port'] = '22'
    os.environ['SERVER.hostname'] = 'demo.com'

    config.load_env()
    assert config['server.port'] == 22
    assert config['server.hostname'] == 'demo.com'

    os.environ['SERVER.PORT'] = '33'
    config.load_env()
    assert config['server.port'] == 33
